// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#pragma once

#include <pthread.h>	// for pthread_t

#include <atomic>							 // for atomic, memory_order_relaxed
#include <cassert>						 // for assert
#include <condition_variable>	 // for condition_variable
#include <cstdint>						 // for intptr_t
#include <cstdlib>						 // for abort
#include <functional>					 // for function
#include <mutex>							 // for mutex, lock_guard, unique_lock
#include <random>
#include <thread>	 // for thread
#include <vector>	 // for vector

#include "Common.hpp"	 // for ALIGN_TO_CACHE_LINE
#include "Debug.hpp"	 // for LogSubsystem, LogSubsystem::RUNTI, Logger
#include "Emper.hpp"	 // for WORKER_NOTIFY
#include "NextFiberResult.hpp"
#include "Scheduler.hpp"	// for Scheduler
#include "Worker.hpp"
#include "emper-common.h"			 // for workerid_t
#include "lib/sync/Latch.hpp"	 // for Latch

class ContextManager;
class Dispatcher;
class Fiber;
class RuntimeStrategy;
class RuntimeStrategyFactory;

namespace emper::io {
class IoContext;
}

using emper::io::IoContext;

enum WakeupMode {
	IF_SLEEPING_OBSERVED,
	ALWAYS,
};

class Runtime : public Logger<LogSubsystem::RUNTI> {
 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	const workerid_t workerCount;

	std::vector<std::function<void(void)>> newWorkerHooks;

	Latch workerLatch;

	RuntimeStrategy* const strategy;
	Scheduler& scheduler;
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;
	Worker** workers;
	IoContext* globalIo = nullptr;
	IoContext* ioContexts;

	std::default_random_engine randomEngine;
	std::uniform_int_distribution<unsigned int> uniformIntDistribution;

	auto workerLoop(Worker* worker) -> void*;

	ALIGN_TO_CACHE_LINE std::mutex workerSleepMutex;
	std::condition_variable workerSleepConditionVariable;
	ALIGN_TO_CACHE_LINE std::atomic<unsigned long> sleepingWorkers;
	bool skipSleep;

	static RuntimeStrategyFactory& DEFAULT_STRATEGY;

	static void printLastRuntimeStats();

 protected:
	void addNewWorkerHook(const std::function<void(void)>& hook) { newWorkerHooks.push_back(hook); };

	template <WakeupMode wakeupMode = WakeupMode::IF_SLEEPING_OBSERVED>
	inline void wakeupSleepingWorkers() {
		if constexpr (wakeupMode == WakeupMode::IF_SLEEPING_OBSERVED) {
			// If we observe no worker sleeping, then we do not try to
			// attempt to wakeup one. Note that this is racy, i.e., just
			// because we do not see any worker sleeping, does not mean that
			// one (or all but us, as a matter of a fact) is about to
			// sleep. However, in the worst case, this causes a worker to
			// sleep longer, until this worker (or another active worker)
			// schedules another fiber, when this method will be called
			// again. And then it is likely that the scheduling worker will
			// observe the sleeping worker.
			if (!sleepingWorkers.load(std::memory_order_relaxed)) {
				return;
			}
		}

		std::lock_guard<std::mutex> lk(workerSleepMutex);
		skipSleep = true;
		if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::all) {
			sleepingWorkers.store(0, std::memory_order_relaxed);
			workerSleepConditionVariable.notify_all();
		} else if constexpr (emper::WORKER_WAKEUP_STRATEGY == emper::WorkerWakeupStrategy::one) {
			sleepingWorkers.fetch_sub(1, std::memory_order_relaxed);
			workerSleepConditionVariable.notify_one();
		} else {
			// Unknown worker wakeup strategy.
			abort();
		}
	}

	void dispatcherLoopSleep() {
		if (skipSleep) {
			skipSleep = false;
			return;
		}

		std::unique_lock<std::mutex> lk(workerSleepMutex);
		// Check again if "skip sleep" has been set.
		if (skipSleep) return;

		sleepingWorkers.fetch_add(1, std::memory_order_relaxed);
		workerSleepConditionVariable.wait(lk);
	}

	inline auto getGlobalIo() -> IoContext* {
		if constexpr (emper::IO) {
			return globalIo;
		}

		return nullptr;
	}

 public:
	Runtime() : Runtime(std::thread::hardware_concurrency()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY) {}

	Runtime(RuntimeStrategyFactory& strategyFactory)
			: Runtime(std::thread::hardware_concurrency(), strategyFactory) {}

	Runtime(workerid_t workerCount, RuntimeStrategyFactory& strategyFactory,
					unsigned int seed = std::random_device()());

	~Runtime();

	inline void schedule(Fiber& fiber) {
		// Calling schedule() only works from within the EMPER runtime.
		assert(inRuntime());

		scheduler.schedule(fiber);
	}

	inline void scheduleFromAnywhere(Fiber& fiber) { scheduler.scheduleFromAnywhere(fiber); }

	// TODO: This should probably not be a public method of Runtime.
	auto nextFiber() -> NextFiberResult;

	// https://stackoverflow.com/a/3747462/194894
	static inline auto rand() -> int { return Worker::rand(); }

	static inline auto getWorkerId() -> workerid_t { return Worker::getCurrentWorkerId(); }

	[[nodiscard]] inline auto getWorkerCount() const -> workerid_t { return workerCount; }

	static inline auto getRuntime() -> Runtime* { return currentRuntime; }

	inline auto getContextManager() -> ContextManager& { return contextManager; }

	inline auto getStrategy() -> RuntimeStrategy& { return *strategy; }

	void waitUntilFinished();

	void printStats();

	static auto inRuntime() -> bool { return Worker::isWorkerThread(); }

	void executeAndWait(std::function<void()> f);

	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	friend IoContext;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
};
