// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <liburing.h>		 // for io_uring_prep_read, io_uring_p...
#include <sys/socket.h>	 // for socklen_t
#include <sys/types.h>	 // for ssize_t

#include <cstddef>	// for size_t
#include <cstdint>	// for int32_t, uint8_t
#include <cstdlib>	// for abort
#include <functional>
#include <ostream>	// for operator<<, ostream, basic_ost...
#include <stdexcept>
#include <string>

#include "BinaryPrivateSemaphore.hpp"	 // for BPS
#include "CallerEnvironment.hpp"			 // for CallerEnvironment, ANYWHERE
#include "Common.hpp"
#include "Debug.hpp"				 // for LOGD, LogSubsystem, LogSubsyst...
#include "Emper.hpp"				 // for DEBUG
#include "io/Operation.hpp"	 // for Operation, operator<<, Operati...

struct io_uring_sqe;

namespace emper::io {
class Stats;

class FutureError : public std::logic_error {
	friend class Future;
	FutureError(const std::string& what) : std::logic_error(what) {}
};

/*
 * @brief Future representing an IO request which can be awaited
 */
class Future : public Logger<LogSubsystem::IO> {
	friend class IoContext;
	friend class Stats;

 public:
	using Callback = std::function<void(const int32_t&)>;

 protected:
	using State = struct State {
		uint8_t submitted : 1 = 0;	/*!< An sqe for this Future was prepared */
		uint8_t completed : 1 = 0;	/*!< The semaphor was signaled  */
		uint8_t retrieved : 1 = 0;	/*!< Future::wait() has retrieved the result */
		uint8_t dependency : 1 = 0; /*!< This Future is a dependency of another */
		uint8_t cancelled : 1 = 0;	/*!< Future::cancel() was called */
		uint8_t forgotten : 1 =
				0; /*!< This Future does not care about the result of the IO operation */
	};

	BPS sem;

	State state;

	/* IO operation to perform */
	const Operation op;

	/* Arguments for the IO operation */
	const int fd;
	void* const buf;
	const size_t len;
	const int offsetOrFlags;

	/**
	 * Return value of the operation.
	 * __s32 is used by liburing for the res field in its cqe struct
	 * This value is returned by Future::wait() and can only differentiate between
	 * failure or full success.
	 */
	int32_t returnValue = 0;

	// Dependency futures will be submitted to the io_uring before
	// this and will be linked to this IO request using IOSQE_IO_LINK
	Future* dependency = nullptr;

	/**
	 * Callback function which is called by a new scheduled Fiber on completion
	 * A new Callback object is heap allocated in setCallback and freed by the
	 * executing Fiber after callback returned;
	 * The heap allocation and thus the Callback pointer is required to decouple the lifetimes
	 * of the Future used for submission and the Callback which must life for the
	 * time the request is in the io_uring.
	 * The callback is called with the value causing the completion
	 */
	Callback* callback = nullptr;

	virtual void prepareSqe(io_uring_sqe* sqe) = 0;

	void setCompletion(int32_t res) {
		LOGD("Complete " << this);
		returnValue = res;
		state.completed = true;
	}

	virtual void complete(int32_t res) {
		setCompletion(res);
		sem.signal();
	}

	virtual void completeFromAnywhere(int32_t res) {
		setCompletion(res);
		sem.signalFromAnywhere();
	}

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	virtual void recordCompletion(Stats& stats, int32_t res) {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

	Future(Operation op, int fd, void* buf, size_t len, int offsetOrFlags)
			: op(op), fd(fd), buf(buf), len(len), offsetOrFlags(offsetOrFlags){};

 public:
	// Clang-tidy warns about the exception possibly thrown by
	// ~Future -> cancel -> wait
	// But this exception will never be thrown because wait() throws an exception
	// only if callback is set and if callback is set ~Future does not call cancel.
	// NOLINTNEXTLINE(bugprone-exception-escape)
	virtual ~Future() {
		if (isForgotten() || callback) {
			return;
		}

		if constexpr (emper::DEBUG) {
			if (!isRetrieved() && !isDependency() && !isCancelled()) {
				LOGE(this << " created but never awaited");
				abort();
			}
		}

		// cancel this Future to guaranty memory safety
		// By canceling we assure that the this pointer and the contained buffer
		// passed to io_uring is not used any more by the kernel or the IoContext
		cancel();
	}

	[[nodiscard]] auto isRetrieved() const -> bool { return state.retrieved; }
	[[nodiscard]] auto isSubmitted() const -> bool { return state.submitted; }
	[[nodiscard]] auto isCompleted() const -> bool { return state.completed; }
	[[nodiscard]] auto isCancelled() const -> bool { return state.cancelled; }
	[[nodiscard]] auto isDependency() const -> bool { return state.dependency; }
	[[nodiscard]] auto isForgotten() const -> bool { return state.forgotten; }

	/*
	 * @brief reset the Future
	 *
	 * Reset the state and the BPS.
	 * Resetting a Future is useful if we wan't to use the same Future object
	 * multiple times.
	 * A Future for a specific reoccurring IO operation can be created and reused
	 * in a submit, reset loop.
	 */
	virtual inline void reset() {
		LOGD("Resetting Future");
		state = State();
		sem.reset();
	}

	/*
	 * @brief set Future as dependency
	 *
	 * The dependency must complete before this future.
	 * This is assured by linking the futures' sqes in the io_uring using the IOSQE_IO_LINK flag.
	 * If a request in the link fails all dependent futures will return -ECANCELED.
	 *
	 * See: <a href="https://unixism.net/loti/tutorial/link_liburing.html">Liburing linking
	 * requests</a>
	 * The emper equivalent to the example from the liburing documentation.
	 *
	 * @code
	 * int fd = open(FILE_NAME, O_RDWR|O_TRUNC|O_CREAT, 0644);
	 * if (fd < 0 ) {
	 * perror("open");
	 * return 1;
	 * }
	 *
	 * char* msg = "Hallo emper!\n";
	 * WriteFuture writeFuture(fd, &msg, strlen(msg), 0);
	 *
	 * char buf[32];
	 * ReadFuture readFuture(fd, &buf, sizeof(buf), 0);
	 * readFuture.setDependency(writeFuture);
	 *
	 * CloseFuture closeFuture(fd);
	 * closeFuture.setDependency(readFuture);
	 *
	 * // Blocks till all three futures are completed
	 * int bytes_read = closeFuture.submitAndWait();
	 *
	 * // So request in the chain could not be fully completed
	 * if (bytes_read == -ECANCELED) {
	 * }
	 * @endcode
	 *
	 */
	inline void setDependency(Future& dependency) {
		dependency.state.dependency = true;
		this->dependency = &dependency;
	}

 private:
	inline void setCallback(Callback* callback) {
		if (unlikely(this->callback)) {
			delete this->callback;
		}

		this->callback = callback;
	}

 public:
	/*
	 * @brief register a callback which is executed in a new Fiber on completion
	 *
	 * @param callback Callback reference which is copied and executed on completion.
	 *                 It gets passed the value causing the completion.
	 */
	inline void setCallback(Callback& callback) {
		/* Call copy constructor */
		setCallback(new Callback(callback));
	}

	/*
	 * @brief register a callback which is executed in a new Fiber on completion
	 *
	 * @param callback Callback rvalue which gets moved and executed on completion.
	 *                 It gets passed the value causing the completion.
	 */
	inline void setCallback(Callback&& callback) {
		/* Call move constructor */
		setCallback(new Callback(callback));
	}

	/*
	 * @brief submit Future for asynchronous completion to the workers IoContext
	 */
	void submit();

	/**
	 * @brief cancel Future and wait if necessary for its cancellation
	 *
	 * @return return the result received from the io_uring
	 */
	auto cancel() -> int32_t;

	/**
	 * @brief Block till the IO request is completed
	 *
	 * @return return the result received from the io_uring
	 */
	auto wait() -> int32_t;

	/**
	 * @brief Block till the IO request is completed and set errno on error
	 *
	 * @return -1 on error and errno will be set, otherwise the return value of the IO request
	 */
	auto waitAndSetErrno() -> ssize_t;

	/**
	 * @brief Equivalent to calling wait() after calling submit()
	 */
	inline auto submitAndWait() -> int32_t {
		submit();
		return wait();
	}

	friend auto operator<<(std::ostream& os, const Future& f) -> std::ostream& {
		return os << f.op << " Future " << (void*)&f;
	}

	friend auto operator<<(std::ostream& os, const Future* f) -> std::ostream& {
		if (f) {
			return os << *f;
		}

		return os << "Future nullptr";
	}
};

class PartialCompletableFuture : public Future {
	friend class Stats;
	friend class IoContext;
	friend class RecvFuture;

	enum class CompletionType {
		Resubmission,
		IncrementalCompletion,
		Completion,
	};

	/**
	 * @brief Try to complete this IO request using the result of an IO operation.
	 *        If the IO request is completed return true otherwise resubmit it.
	 *
	 * @param res Result from an IO operation from a io_uring_cqe or a common syscall
	 *
	 * @return the type of completion
	 */
	template <CallerEnvironment callerEnvironment>
	auto tryComplete(int32_t res) -> CompletionType;

 protected:
	/**
	 * This member is used internally to handle partial completion.
	 * If initialized with -1 it signals that the IO request should immediately return to the user.
	 * A value >= 0 signals that this IO request should not * return to the user until it is
	 * fully completed or a persistent error arises.
	 * The IO request is considered fully completed if partialCompletion == len.
	 *
	 * When the IO request returns to the user the combination of @c future.returnValue and
	 * @c future.partialCompletion can signal partial completion.
	 */
	ssize_t partialCompletion;

	// return immediately to the user
	static const ssize_t DISABLE_PARTIAL_COMPLETION = -1;
	// try to fully complete the request
	static const ssize_t ENABLE_PARTIAL_COMPLETION = 0;

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, int offsetOrFlags,
													 bool completeFully)
			: Future(op, fd, buf, len, offsetOrFlags),
				partialCompletion(completeFully ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION){};

	PartialCompletableFuture(Operation op, int fd, void* buf, size_t len, int offsetOrFlags,
													 ssize_t partialCompletion)
			: Future(op, fd, buf, len, offsetOrFlags), partialCompletion(partialCompletion){};

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	void recordCompletion(Stats& stats, int32_t res) override {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

	void complete(int32_t res) override {
		CompletionType completion = tryComplete<CallerEnvironment::EMPER>(res);

		switch (completion) {
			case CompletionType::Resubmission:
				return;
			case CompletionType::IncrementalCompletion:
				Future::complete(partialCompletion);
				return;
			case CompletionType::Completion:
				Future::complete(res);
				return;
		}
	}

	void completeFromAnywhere(int32_t res) override {
		CompletionType completion = tryComplete<CallerEnvironment::ANYWHERE>(res);

		switch (completion) {
			case CompletionType::Resubmission:
				return;
			case CompletionType::IncrementalCompletion:
				Future::completeFromAnywhere(partialCompletion);
				return;
			case CompletionType::Completion:
				Future::completeFromAnywhere(res);
				return;
		}
	}

 public:
	/**
	 * @brief Block till the IO request is completed
	 *
	 * @param[out] partialCompletion The amount of work done till an persistent error arose
	 *
	 * @return return the result received from the io_uring
	 */
	auto wait(ssize_t& partialCompletion) -> int32_t {
		int32_t res = Future::wait();
		partialCompletion = this->partialCompletion;
		return res;
	}
	using Future::wait;

	/*
	 * @brief Equivalent to calling wait(partialCompletion) after calling submit()
	 */
	inline auto submitAndWait(ssize_t& partialCompletion) -> int32_t {
		Future::submit();
		return wait(partialCompletion);
	}
	using Future::submitAndWait;

	/**
	 * @brief reset the Future and its partialCompletion.
	 */
	void reset() override {
		if (partialCompletion != DISABLE_PARTIAL_COMPLETION) {
			partialCompletion = ENABLE_PARTIAL_COMPLETION;
		}

		Future::reset();
	}
};

class SendFuture : public PartialCompletableFuture {
	void prepareSqe(io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_send(sqe, fd, buf, len, offsetOrFlags);
		} else {
			io_uring_prep_send(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion,
												 offsetOrFlags);
		}
	}

 public:
	SendFuture(int socket, const void* buffer, size_t length, int flags, ssize_t partialCompletion)
			: PartialCompletableFuture(Operation::SEND, socket, const_cast<void*>(buffer), length, flags,
																 partialCompletion){};

	SendFuture(int socket, const void* buffer, size_t length, int flags, bool send_all = true)
			: SendFuture(socket, const_cast<void*>(buffer), length, flags,
									 send_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION){};
};

class RecvFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_recv(sqe, fd, buf, len, offsetOrFlags);
	}

 public:
	RecvFuture(int socket, void* buffer, size_t length, int flags)
			: Future(Operation::RECV, socket, buffer, length, flags){};
};

class ConnectFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_connect(sqe, fd, (const struct sockaddr*)buf, (socklen_t)len);
	}

 public:
	ConnectFuture(int socket, const struct sockaddr* address, socklen_t address_len)
			: Future(Operation::CONNECT, socket, (void*)address, address_len, 0){};
};

class AcceptFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_accept(sqe, fd, (struct sockaddr*)buf, (socklen_t*)len, 0);
	}

 public:
	AcceptFuture(int socket, const struct sockaddr* address, socklen_t* address_len)
			: Future(Operation::ACCEPT, socket, (void*)address, (size_t)address_len, 0){};
};

class ReadFuture : public PartialCompletableFuture {
	void prepareSqe(io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_read(sqe, fd, buf, len, offsetOrFlags);
		} else {
			io_uring_prep_read(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion,
												 offsetOrFlags + partialCompletion);
		}
	}

 public:
	ReadFuture(int fildes, void* buf, size_t nbyte, int offset, bool read_all = false)
			: PartialCompletableFuture(
						Operation::READ, fildes, buf, nbyte, offset,
						read_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION){};
};

class OpenatFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_openat(sqe, fd, reinterpret_cast<const char*>(buf), len, offsetOrFlags);
	}

 public:
	OpenatFuture(int dirfd, const void* pathname, int flags, mode_t mode = 0)
			: Future(Operation::OPENAT, dirfd, const_cast<void*>(pathname), flags, mode){};
};

class WriteFuture : public PartialCompletableFuture {
	void prepareSqe(io_uring_sqe* sqe) override {
		if (partialCompletion == DISABLE_PARTIAL_COMPLETION) {
			io_uring_prep_write(sqe, fd, buf, len, offsetOrFlags);
		} else {
			// TODO: think about partial writes with incremental offset!
			// Are transparent incremental writes on files without O_APPEND even reasonable?
			io_uring_prep_write(sqe, fd, (char*)buf + partialCompletion, len - partialCompletion,
													offsetOrFlags + partialCompletion);
		}
	}

 public:
	WriteFuture(int fildes, const void* buf, size_t nbyte, int offset, bool write_all = true)
			: PartialCompletableFuture(
						Operation::WRITE, fildes, const_cast<void*>(buf), nbyte, offset,
						write_all ? ENABLE_PARTIAL_COMPLETION : DISABLE_PARTIAL_COMPLETION){};
};

class WritevFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_writev(sqe, fd, (const struct iovec*)buf, len, 0);
	}

	/**
	 * Used for Stats::recordCompletion double dispatch
	 */
	void recordCompletion(Stats& stats, int32_t res) override {
		if constexpr (emper::STATS) {
			recordCompletionInternal(stats, res);
		}
	}

	void recordCompletionInternal(Stats& stats, int32_t res) const;

 public:
	WritevFuture(int fildes, const struct iovec* iov, int iovnct)
			: Future(Operation::WRITEV, fildes, (void*)iov, iovnct, 0){};
};

class CloseFuture : public Future {
	void prepareSqe(io_uring_sqe* sqe) override { io_uring_prep_close(sqe, fd); }

 public:
	CloseFuture(int fildes) : Future(Operation::CLOSE, fildes, nullptr, 0, 0){};

	void submitAndForget() {
		state.forgotten = true;
		submit();
	}
};

/*
 * @brief Add a timeout to any Future not already submitted.
 *
 * A Timeout is added to a Future by inserting an dependent IOURING_OP_LINK_TIMEOUT
 * sqe after the actual IO operation.
 * The dependent timeout will not be started like usual dependent requests after
 * the previous request is completed, it is instead armed immediately when the
 * actual request is started.
 * See: https://lwn.net/Articles/803932/
 *
 * @code
 * ReadFuture rf(fd, buf, buf_len, 0);
 * struct __kernel_timespec ts = {.tv_sec = 0, .tv_nsec = 0);
 * TimeoutWrapper t(rf, ts);
 * // Both futures should be awaited by calling its wait method to
 * // prevent use after free problems.
 * t.submitAndWait()
 * rf.Wait()
 * @endcode
 */
class TimeoutWrapper : public Future {
 public:
	using Timespec = struct __kernel_timespec;

	TimeoutWrapper(Future& future, Timespec& ts)
			: Future(Operation::LINK_TIMEOUT, 0, (void*)&ts, 0, 0) {
		setDependency(future);
	};

 private:
	void prepareSqe(io_uring_sqe* sqe) override {
		io_uring_prep_link_timeout(sqe, (Timespec*)buf, 0);
	}
};

/*
 * @brief Arm a timeout which will signal the future when it is reached
 *
 * If the timeout was reached calling wait() will return -ETIME.
 */
class AlarmFuture : public Future {
 public:
	using Timespec = struct __kernel_timespec;

	AlarmFuture(Timespec& ts) : Future(Operation::TIMEOUT, 0, (void*)&ts, 0, 0){};

 private:
	void prepareSqe(io_uring_sqe* sqe) override { io_uring_prep_timeout(sqe, (Timespec*)buf, 0, 0); }
};

/*
 * @brief Request the cancellation of another Future and await the cancellation
 *
 * If the target was found and cancelled wait() will immediately return 0.
 * If the target was already completed wait() will immediately return -ENOENT.
 * If the target was already started wait() will return -EALREADY.
 */
class CancelWrapper : public Future {
	void prepareSqe(io_uring_sqe* sqe) override { io_uring_prep_cancel(sqe, buf, 0); }

 public:
	CancelWrapper(Future& future) : Future(Operation::CANCEL, 0, &future, 0, 0){};
};
}	 // namespace emper::io
