// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#include "LawsDispatcher.hpp"

#include <atomic>	 // for atomic, memory_order_relaxed

#include "Common.hpp"	 // for DIE_MSG
#include "Emper.hpp"
#include "LawsStrategy.hpp"	 // for LawsStrategy, LawsStrategy::FiberSource
#include "NextFiberResult.hpp"
#include "Runtime.hpp"

void LawsDispatcher::recycle(Fiber* fiber) {
	// If the ref count has not reached zero yet, do not recycle the
	// fiber.
	if (decreaseRefCount(fiber)) return;

	Dispatcher::recycle(fiber);
}

void LawsDispatcher::dispatchLoop() {
	while (true) {
		NextFiberResult next = runtime.nextFiber();
		Fiber* const fiber = next.fiber;
		if (!fiber) {
			dispatchLoopDoSleep();

			continue;
		}

		// The isRunnable() method uses an exchange on an atomic
		// boolean initialize to true in order to check if this fiber
		// is runnable.
		if (isRunnable(fiber)) {
			if constexpr (emper::STATS) {
				auto fiberSource = static_cast<LawsStrategy::FiberSource>(next.metadata);
				switch (fiberSource) {
					case LawsStrategy::FiberSource::fromPriority:
						lawsStrategy.dispatchedFiberFromPriority.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::local:
						lawsStrategy.dispatchedFiberFromLocal.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::stolen:
						lawsStrategy.dispatchedFiberStolen.fetch_add(1, std::memory_order_relaxed);
						break;
					case LawsStrategy::FiberSource::anywhereQueue:
						lawsStrategy.dispatchedFiberFromAnywhere.fetch_add(1, std::memory_order_relaxed);
						break;
					default:
						DIE_MSG("Unknown fiber source: " << next.metadata);
						break;
				}
			}

			// The fiber was marked das runnable. Run it now.
			dispatch(fiber);
			// Update the affinity if one was set.
			workeraffinity_t* const affinity = getAffinityBuffer(fiber);
			if (affinity) *affinity = Runtime::getWorkerId();
		}

		recycle(fiber);
	}
}
